# pylint: disable=missing-class-docstring,missing-function-docstring
"""pure python trajectory calculation backend"""

import math
from dataclasses import dataclass
from typing import NamedTuple

from .conditions import Atmo, Shot, Wind
from .munition import Ammo, Weapon
from .settings import Settings
from .trajectory_data import TrajectoryData, TrajFlag
from .unit import Distance, Angular, Velocity, Weight, Energy, Pressure, Temperature

__all__ = ('TrajectoryCalc', )

cZeroFindingAccuracy = 0.000005
cMinimumVelocity = 50.0
cMaximumDrop = -15000
cMaxIterations = 20
cGravityConstant = -32.17405


class CurvePoint(NamedTuple):
    a: float
    b: float
    c: float


@dataclass
class Vector:
    x: float
    y: float
    z: float

    def magnitude(self):
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def mul_by_const(self, a: float):
        return Vector(self.x * a, self.y * a, self.z * a)

    def mul_by_vector(self, b: 'Vector'):
        return self.x * b.x + self.y * b.y + self.z * b.z

    def add(self, b: 'Vector'):
        return Vector(self.x + b.x, self.y + b.y, self.z + b.z)

    def subtract(self, b: 'Vector'):
        return Vector(self.x - b.x, self.y - b.y, self.z - b.z)

    def negate(self):
        return Vector(-self.x, -self.y, -self.z)

    def normalize(self):
        m = self.magnitude()
        if math.fabs(m) < 1e-10:
            return Vector(self.x, self.y, self.z)
        return self.mul_by_const(1.0 / m)

    def __add__(self, other: 'Vector'):
        return self.add(other)

    def __radd__(self, other: 'Vector'):
        return self.add(other)

    def __iadd__(self, other: 'Vector'):
        return self.add(other)

    def __sub__(self, other: 'Vector'):
        return self.subtract(other)

    def __rsub__(self, other: 'Vector'):
        return self.subtract(other)

    def __isub__(self, other: 'Vector'):
        return self.subtract(other)

    def __mul__(self, other: [int, float, 'Vector']):
        if isinstance(other, (int, float)):
            return self.mul_by_const(other)
        if isinstance(other, Vector):
            return self.mul_by_vector(other)
        raise TypeError(other)

    def __rmul__(self, other: [int, float, 'Vector']):
        return self.__mul__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self.negate()


class TrajectoryCalc:

    def __init__(self, ammo: Ammo):
        self.ammo = ammo
        self._bc = self.ammo.dm.value
        self._table_data = ammo.dm.drag_table
        self._curve = calculate_curve(self._table_data)

    def get_calc_step(self, step: float):

        maximum_step = Settings._MAX_CALC_STEP_SIZE
        step /= 2

        if step > maximum_step:
            step_order = int(math.floor(math.log10(step)))
            maximum_order = int(math.floor(math.log10(maximum_step)))
            step /= math.pow(10, step_order - maximum_order + 1)

        return step

    def zero_angle(self, weapon: Weapon, atmo: Atmo):
        return self._zero_angle(self.ammo, weapon, atmo)

    def trajectory(self, weapon: Weapon, shot_info: Shot, step: [float, Distance],
                   extra_data: bool = False):

        dist_step = Settings.Units.distance(step)
        atmo = shot_info.atmo
        winds = shot_info.winds
        filter_flags = TrajFlag.RANGE

        if extra_data:
            print('ext', extra_data)
            dist_step = Distance.Foot(0.2)
            filter_flags = TrajFlag.ALL
        return self._trajectory(self.ammo, weapon, atmo, shot_info, winds, dist_step, filter_flags)

    def _zero_angle(self, ammo: Ammo, weapon: Weapon, atmo: Atmo):
        calc_step = self.get_calc_step(weapon.zero_distance.units(10) >> Distance.Foot)
        zero_distance = math.cos(
            weapon.zero_look_angle >> Angular.Radian
        ) * (weapon.zero_distance >> Distance.Foot)
        height_at_zero = math.sin(
            weapon.zero_look_angle >> Angular.Radian
        ) * (weapon.zero_distance >> Distance.Foot)
        maximum_range = zero_distance + calc_step
        sight_height = weapon.sight_height >> Distance.Foot
        mach = atmo.mach >> Velocity.FPS
        density_factor = atmo.density_factor()
        muzzle_velocity = ammo.mv >> Velocity.FPS
        barrel_azimuth = 0.0
        barrel_elevation = math.atan(height_at_zero / zero_distance)
        iterations_count = 0
        zero_finding_error = cZeroFindingAccuracy * 2
        gravity_vector = Vector(.0, cGravityConstant, .0)

        # x - distance towards target, y - drop and z - windage
        while zero_finding_error > cZeroFindingAccuracy and iterations_count < cMaxIterations:
            velocity = muzzle_velocity
            time = 0.0
            range_vector = Vector(.0, -sight_height, .0)
            velocity_vector = Vector(
                math.cos(barrel_elevation) * math.cos(barrel_azimuth),
                math.sin(barrel_elevation),
                math.cos(barrel_elevation) * math.sin(barrel_azimuth)
            ) * velocity

            while range_vector.x <= maximum_range:
                if velocity < cMinimumVelocity or range_vector.y < cMaximumDrop:
                    break

                delta_time = calc_step / velocity_vector.x

                drag = density_factor * velocity * self.drag_by_mach(velocity / mach)

                velocity_vector -= (velocity_vector * drag - gravity_vector) * delta_time
                delta_range_vector = Vector(calc_step, velocity_vector.y * delta_time,
                                            velocity_vector.z * delta_time)
                range_vector += delta_range_vector
                velocity = velocity_vector.magnitude()
                time += delta_range_vector.magnitude() / velocity

                if math.fabs(range_vector.x - zero_distance) < 0.5 * calc_step:
                    zero_finding_error = math.fabs(range_vector.y - height_at_zero)
                    if zero_finding_error > cZeroFindingAccuracy:
                        barrel_elevation -= (range_vector.y - height_at_zero) / range_vector.x
                    break

            iterations_count += 1

        return Angular.Radian(barrel_elevation)

    def _trajectory(self, ammo: Ammo, weapon: Weapon, atmo: Atmo,
                    shot_info: Shot, winds: list[Wind],
                    dist_step: Distance, filter_flags: TrajFlag):

        time = 0
        look_angle = weapon.zero_look_angle >> Angular.Radian
        twist = weapon.twist >> Distance.Inch
        length = ammo.length >> Distance.Inch
        diameter = ammo.dm.diameter >> Distance.Inch
        weight = ammo.dm.weight >> Weight.Grain

        # step = shot_info.step >> Distance.Foot
        step = dist_step >> Distance.Foot
        calc_step = self.get_calc_step(step)

        maximum_range = (shot_info.max_range >> Distance.Foot) + 1

        ranges_length = int(maximum_range / step) + 1
        len_winds = len(winds)
        current_wind = 0
        current_item = 0

        stability_coefficient = 1.0
        next_wind_range = 1e7

        barrel_elevation = (shot_info.zero_angle >> Angular.Radian) + (
                shot_info.relative_angle >> Angular.Radian)
        alt0 = atmo.altitude >> Distance.Foot
        sight_height = weapon.sight_height >> Distance.Foot

        next_range_distance = .0
        barrel_azimuth = .0
        previous_mach = .0

        gravity_vector = Vector(.0, cGravityConstant, .0)
        range_vector = Vector(.0, -sight_height, .0)

        ranges = []

        if len_winds < 1:
            wind_vector = Vector(.0, .0, .0)
        else:
            if len_winds > 1:
                next_wind_range = winds[0].until_distance() >> Distance.Foot
            wind_vector = wind_to_vector(shot_info, winds[0])

        if Settings.USE_POWDER_SENSITIVITY:
            velocity = ammo.get_velocity_for_temp(atmo.temperature) >> Velocity.FPS
        else:
            velocity = ammo.mv >> Velocity.FPS

        # x - distance towards target, y - drop and z - windage
        velocity_vector = Vector(math.cos(barrel_elevation) * math.cos(barrel_azimuth),
                                 math.sin(barrel_elevation),
                                 math.cos(barrel_elevation) * math.sin(barrel_azimuth)) * velocity

        if twist != 0 and length and diameter:
            stability_coefficient = calculate_stability_coefficient(ammo, weapon, atmo)
            twist_coefficient = -1 if twist > 0 else 1

        # With non-zero look_angle, rounding can suggest multiple adjacent zero-crossings
        seen_zero = TrajFlag.NONE  # Record when we see each zero crossing so we only register one
        if range_vector.y >= 0:
            seen_zero |= TrajFlag.ZERO_UP  # We're starting above zero; we can only go down
        elif range_vector.y < 0 and barrel_elevation < look_angle:
            seen_zero |= TrajFlag.ZERO_DOWN  # We're below and pointing down from look angle; no zeroes!

        while range_vector.x <= maximum_range + calc_step:
            _flag = TrajFlag.NONE

            if velocity < cMinimumVelocity or range_vector.y < cMaximumDrop:
                break

            density_factor, mach = atmo.get_density_factor_and_mach_for_altitude(
                alt0 + range_vector.y)

            if range_vector.x >= next_wind_range:
                current_wind += 1
                wind_vector = wind_to_vector(shot_info, winds[current_wind])

                if current_wind == len_winds - 1:
                    next_wind_range = 1e7
                else:
                    next_wind_range = winds[current_wind].until_distance >> Distance.Foot

            # Zero-crossing checks
            if range_vector.x > 0:
                # Zero reference line is the sight line defined by look_angle
                reference_height = range_vector.x * math.tan(look_angle)
                # If we haven't seen ZERO_UP, we look for that first
                if not seen_zero & TrajFlag.ZERO_UP:
                    if range_vector.y >= reference_height:
                        _flag |= TrajFlag.ZERO_UP
                        seen_zero |= TrajFlag.ZERO_UP
                # We've crossed above sight line; now look for crossing back through it
                elif not seen_zero & TrajFlag.ZERO_DOWN:
                    if range_vector.y < reference_height:
                        _flag |= TrajFlag.ZERO_DOWN
                        seen_zero |= TrajFlag.ZERO_DOWN

            # Mach crossing check
            if (velocity / mach <= 1) and (previous_mach > 1):
                _flag |= TrajFlag.MACH

            # Next range check
            if range_vector.x >= next_range_distance:
                _flag |= TrajFlag.RANGE
                next_range_distance += step
                current_item += 1

            if _flag & filter_flags:

                windage = range_vector.z

                if twist != 0:
                    windage += (1.25 * (stability_coefficient + 1.2)
                                * math.pow(time, 1.83) * twist_coefficient) / 12

                ranges.append(create_trajectory_row(
                    time, range_vector, velocity_vector,
                    velocity, mach, windage, weight, _flag.value
                ))

                if current_item == ranges_length:
                    break

            previous_mach = velocity / mach

            velocity_adjusted = velocity_vector - wind_vector

            delta_time = calc_step / velocity_vector.x
            velocity = velocity_adjusted.magnitude()

            drag = density_factor * velocity * self.drag_by_mach(velocity / mach)

            velocity_vector -= (velocity_adjusted * drag - gravity_vector) * delta_time
            delta_range_vector = Vector(calc_step,
                                        velocity_vector.y * delta_time,
                                        velocity_vector.z * delta_time)
            range_vector += delta_range_vector
            velocity = velocity_vector.magnitude()
            time += delta_range_vector.magnitude() / velocity

        return ranges

    def drag_by_mach(self, mach: float):
        cd = calculate_by_curve(self._table_data, self._curve, mach)
        return cd * 2.08551e-04 / self._bc

    @property
    def cdm(self):
        return self._cdm()

    def _cdm(self):
        """
        Returns custom drag function based on input data
        """
        drag_table = self.ammo.dm.drag_table
        cdm = []
        bc = self.ammo.dm.value

        for point in drag_table:
            st_mach = point['Mach']
            st_cd = calculate_by_curve(drag_table, self._curve, st_mach)
            cd = st_cd * bc
            cdm.append({'CD': cd, 'Mach': st_mach})

        return cdm


def calculate_stability_coefficient(ammo: Ammo, rifle: Weapon, atmo: Atmo):
    weight = ammo.dm.weight >> Weight.Grain
    diameter = ammo.dm.diameter >> Distance.Inch
    twist = math.fabs(rifle.twist >> Distance.Inch) / diameter
    length = (ammo.length >> Distance.Inch) / diameter
    ft = atmo.temperature >> Temperature.Fahrenheit
    mv = ammo.mv >> Velocity.FPS
    pt = atmo.pressure >> Pressure.InHg
    sd = 30 * weight / (
            math.pow(twist, 2) * math.pow(diameter, 3) * length * (1 + math.pow(length, 2))
    )
    fv = math.pow(mv / 2800, 1.0 / 3.0)
    ftp = ((ft + 460) / (59 + 460)) * (29.92 / pt)
    return sd * fv * ftp


def wind_to_vector(shot: Shot, wind: Wind):
    sight_cosine = math.cos(shot.zero_angle >> Angular.Radian)
    sight_sine = math.sin(shot.zero_angle >> Angular.Radian)
    cant_cosine = math.cos(shot.cant_angle >> Angular.Radian)
    cant_sine = math.sin(shot.cant_angle >> Angular.Radian)
    range_velocity = (wind.velocity >> Velocity.FPS) * math.cos(
        wind.direction_from >> Angular.Radian)
    cross_component = (wind.velocity >> Velocity.FPS) * math.sin(
        wind.direction_from >> Angular.Radian)
    range_factor = -range_velocity * sight_sine
    return Vector(range_velocity * sight_cosine,
                  range_factor * cant_cosine + cross_component * cant_sine,
                  cross_component * cant_cosine - range_factor * cant_sine)


def create_trajectory_row(time: float, range_vector: Vector, velocity_vector: Vector,
                          velocity: float, mach: float, windage: float, weight: float, flag: int):
    drop_adjustment = get_correction(range_vector.x, range_vector.y)
    windage_adjustment = get_correction(range_vector.x, windage)
    trajectory_angle = math.atan(velocity_vector.y / velocity_vector.x)


    return TrajectoryData(
        time=time,
        distance=Distance.Foot(range_vector.x),
        drop=Distance.Foot(range_vector.y),
        drop_adj=Angular.Radian(drop_adjustment),
        windage=Distance.Foot(windage),
        windage_adj=Angular.Radian(windage_adjustment),
        velocity=Velocity.FPS(velocity),
        mach=velocity / mach,
        energy=Energy.FootPound(calculate_energy(weight, velocity)),
        angle=Angular.Radian(trajectory_angle),
        ogw=Weight.Pound(calculate_ogv(weight, velocity)),
        flag=flag
    )


def get_correction(distance: float, offset: float):
    if distance != 0:
        return math.atan(offset / distance)
    return 0  # better None


def calculate_energy(bullet_weight: float, velocity: float):
    return bullet_weight * math.pow(velocity, 2) / 450400


def calculate_ogv(bullet_weight: float, velocity: float):
    return math.pow(bullet_weight, 2) * math.pow(velocity, 3) * 1.5e-12


def calculate_curve(data_points):
    # rate, x1, x2, x3, y1, y2, y3, a, b, c
    # curve = []
    # curve_point
    # num_points, len_data_points, len_data_range

    rate = (data_points[1]['CD'] - data_points[0]['CD']) \
           / (data_points[1]['Mach'] - data_points[0]['Mach'])
    curve = [CurvePoint(0, rate, data_points[0]['CD'] - data_points[0]['Mach'] * rate)]
    len_data_points = int(len(data_points))
    len_data_range = len_data_points - 1

    for i in range(1, len_data_range):
        x1 = data_points[i - 1]['Mach']
        x2 = data_points[i]['Mach']
        x3 = data_points[i + 1]['Mach']
        y1 = data_points[i - 1]['CD']
        y2 = data_points[i]['CD']
        y3 = data_points[i + 1]['CD']
        a = ((y3 - y1) * (x2 - x1) - (y2 - y1) * (x3 - x1)) / (
                (x3 * x3 - x1 * x1) * (x2 - x1) - (x2 * x2 - x1 * x1) * (x3 - x1))
        b = (y2 - y1 - a * (x2 * x2 - x1 * x1)) / (x2 - x1)
        c = y1 - (a * x1 * x1 + b * x1)
        curve_point = CurvePoint(a, b, c)
        curve.append(curve_point)

    num_points = len_data_points
    rate = (data_points[num_points - 1]['CD'] - data_points[num_points - 2]['CD']) / \
           (data_points[num_points - 1]['Mach'] - data_points[num_points - 2]['Mach'])
    curve_point = CurvePoint(
        0, rate, data_points[num_points - 1]['CD'] - data_points[num_points - 2]['Mach'] * rate
    )
    curve.append(curve_point)
    return curve


def calculate_by_curve(data: list, curve: list, mach: float):
    """returning the calculated drag for a
    specified mach based on previously calculated data"""
    # num_points, mlo, mhi, mid
    # cdef CurvePoint curve_m

    num_points = int(len(curve))
    mlo = 0
    mhi = num_points - 2

    while mhi - mlo > 1:
        mid = int(math.floor(mhi + mlo) / 2.0)
        if data[mid]['Mach'] < mach:
            mlo = mid
        else:
            mhi = mid

    if data[mhi]['Mach'] - mach > mach - data[mlo]['Mach']:
        m = mlo
    else:
        m = mhi
    curve_m = curve[m]
    return curve_m.c + mach * (curve_m.b + curve_m.a * mach)



"""Default logger for py_ballisticcalc library"""

import logging

__all__ = ('logger',)

formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)

logger = logging.getLogger('py_balcalc')
logger.addHandler(console_handler)
logger.setLevel(logging.INFO)



import unittest
from py_ballisticcalc.unit import *


def back_n_forth(test, value, units):
    u = test.unit_class(value, units)
    v = u >> units
    test.assertAlmostEqual(v, value, 7, f'Read back failed for {units}')


class TestAngular(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Angular
        self.unit_list = [
            Angular.Degree,
            Angular.MOA,
            Angular.MRad,
            Angular.Mil,
            Angular.Radian,
            Angular.Thousand
        ]

    def test_angular(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestDistance(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Distance
        self.unit_list = [
            Distance.Centimeter,
            Distance.Foot,
            Distance.Inch,
            Distance.Kilometer,
            Distance.Line,
            Distance.Meter,
            Distance.Millimeter,
            Distance.Mile,
            Distance.NauticalMile,
            Distance.Yard
        ]

    def test_distance(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestEnergy(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Energy
        self.unit_list = [
            Energy.FootPound,
            Energy.Joule
        ]

    def test_energy(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestPressure(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Pressure
        self.unit_list = [
            Pressure.Bar,
            Pressure.HP,
            Pressure.MmHg,
            Pressure.InHg
        ]

    def test_pressure(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestTemperature(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Temperature
        self.unit_list = [
            Temperature.Fahrenheit,
            Temperature.Kelvin,
            Temperature.Celsius,
            Temperature.Rankin
        ]

    def test_temperature(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestVelocity(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Velocity
        self.unit_list = [
            Velocity.FPS,
            Velocity.KMH,
            Velocity.KT,
            Velocity.MPH,
            Velocity.MPS
        ]

    def test_velocity(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestWeight(unittest.TestCase):

    def setUp(self) -> None:
        self.unit_class = Weight
        self.unit_list = [
            Weight.Grain,
            Weight.Gram,
            Weight.Kilogram,
            Weight.Newton,
            Weight.Ounce,
            Weight.Pound
        ]

    def test_weight(self):
        for u in self.unit_list:
            with self.subTest(unit=u):
                back_n_forth(self, 3, u)


class TestUnitConversionSyntax(unittest.TestCase):

    def setUp(self) -> None:
        self.low = Distance.Yard(10)
        self.high = Distance.Yard(100)

    def test__eq__(self):
        self.assertEqual(self.low, 360)
        self.assertEqual(360, self.low)
        self.assertEqual(self.low, self.low)
        self.assertEqual(self.low, Distance.Foot(30))

    def test__ne__(self):
        self.assertNotEqual(Distance.Yard(100), Distance.Yard(90))

    def test__lt__(self):
        self.assertLess(self.low, self.high)
        self.assertLess(10, self.high)
        self.assertLess(self.low, 9999)

    def test__gt__(self):
        self.assertGreater(self.high, self.low)
        self.assertGreater(self.high, 10)
        self.assertGreater(9000, self.low)

    def test__ge__(self):
        self.assertGreaterEqual(self.high, self.low)
        self.assertGreaterEqual(self.high, self.high)

        self.assertGreaterEqual(self.high, 90)
        self.assertGreaterEqual(self.high, 0)

    def test__le__(self):
        self.assertLessEqual(self.low, self.high)
        self.assertLessEqual(self.high, self.high)

        self.assertLessEqual(self.low, 360)
        self.assertLessEqual(self.low, 360)

    def test__rshift__(self):
        self.assertIsInstance(self.low >> Distance.Meter, (int, float))
        self.low >>= Distance.Meter
        self.assertIsInstance(self.low, (int, float))

    def test__lshift__(self):
        desired_unit_type = Distance
        desired_units = Distance.Foot
        converted = self.low << desired_units
        self.assertIsInstance(converted, desired_unit_type)
        self.assertEqual(converted.units, desired_units)
        self.low <<= desired_units
        self.assertEqual(self.low.units, desired_units)



"""Searching for an available backends"""

from .logger import logger

# trying to use cython based backend
try:
    from py_ballisticcalc_exts import *  # pylint: disable=wildcard-import

    logger.info("Binary modules found, running in binary mode")
except ImportError as error:
    from .drag_model import *
    from .trajectory_calc import *

    logger.warning("Library running in pure python mode. "
                   "For better performance install 'py_ballisticcalc.exts' package")



"""Module to create custom drag function based on Multiple ballistic coefficients"""
import typing
from math import pow as math_pow
from typing import NamedTuple, Iterable

from .conditions import Atmo
# pylint: disable=import-error,no-name-in-module
from .backend import make_data_points
from .settings import Settings as Set
from .unit import Distance, Weight, Velocity

__all__ = ('MultiBC', )


class MultiBCRow(NamedTuple):
    """Multi-BC point, for internal usage"""
    BC: float
    V: Set.Units.velocity


class DragTableRow(NamedTuple):
    """CDM point, for internal usage"""
    CD: float
    Mach: float


class BCMachRow(NamedTuple):
    """BC-Mach point, for internal usage"""
    BC: float
    Mach: float


class MultiBC:  # pylint: disable=too-few-public-methods
    """Creates instance to calculate custom drag tabel based on input multi-bc table"""

    def __init__(self, drag_table: Iterable[dict], diameter: Distance, weight: Weight,
                 mbc_table: Iterable[dict]):

        self.mbc_table = mbc_table
        self.weight = weight
        self.diameter = diameter
        self.sectional_density = self._get_sectional_density()

        atmosphere = Atmo.icao()

        altitude = Distance.Meter(0) >> Distance.Foot
        _, mach = atmosphere.get_density_factor_and_mach_for_altitude(altitude)
        self.speed_of_sound = Velocity.FPS(mach) >> Velocity.MPS

        self._table_data = make_data_points(drag_table)
        self._bc_table = self._parse_mbc(mbc_table)

    def _parse_mbc(self, mbc_table):
        table = []
        for p in mbc_table:
            print(p['V'], Set.Units.velocity)
            print(Set.Units.velocity(p['V']))
            v = Set.Units.velocity(p['V']) >> Velocity.MPS
            mbc = MultiBCRow(p['BC'], v)
            table.append(mbc)
        return sorted(table, reverse=True)

    def _get_sectional_density(self) -> float:
        w = self.weight >> Weight.Grain
        d = self.diameter >> Distance.Inch
        return w / math_pow(d, 2) / 7000

    def _get_form_factor(self, bc) -> float:
        return self.sectional_density / bc

    @staticmethod
    def _get_counted_cd(form_factor, standard_cd):
        return standard_cd * form_factor

    def _interpolate_bc_table(self) -> typing.Generator:
        """
        Extends input bc table by creating bc value for each point of standard Drag Model
        """
        bc_table = tuple(self._bc_table)
        bc_mah = [BCMachRow(bc_table[0].BC, self._table_data[-1].Mach)]
        bc_mah.extend(
            [BCMachRow(point.BC, point.V / self.speed_of_sound) for point in bc_table]
        )
        bc_mah.append(BCMachRow(bc_mah[-1].BC, self._table_data[0].Mach))

        yield bc_mah[0].BC

        for bc_max, bc_min in zip(bc_mah, bc_mah[1:]):
            df_part = [
                point for point in self._table_data if bc_max.Mach > point.Mach >= bc_min.Mach
            ]
            ddf = len(df_part)
            bc_delta = (bc_max.BC - bc_min.BC) / ddf
            for j in range(ddf):
                yield bc_max.BC - bc_delta * j

    def _cdm_generator(self) -> typing.Generator:
        bc_extended = reversed(list(self._interpolate_bc_table()))
        form_factors = [self._get_form_factor(bc) for bc in bc_extended]

        for i, point in enumerate(self._table_data):
            cd = form_factors[i] * point.CD
            yield {'CD': cd, 'Mach': point.Mach}

    @property
    def cdm(self) -> list[dict]:
        """
        :return: custom drag function based on input multiple ballistic coefficients
        """
        return list(self._cdm_generator())



"""
Use-full types for units of measurement conversion for ballistics calculations
"""

import typing
from enum import IntEnum
from math import pi, atan, tan
from typing import NamedTuple
from dataclasses import dataclass

__all__ = ('Unit', 'AbstractUnit', 'UnitPropsDict', 'Distance',
           'Velocity', 'Angular', 'Temperature', 'Pressure',
           'Energy', 'Weight', 'TypedUnits')


class Unit(IntEnum):
    """
    Usage of IntEnum simplify data serializing for using it with databases etc.
    """
    RAD = 0
    DEGREE = 1
    MOA = 2
    MIL = 3
    MRAD = 4
    THOUSAND = 5
    INCHES_PER_100YD = 6
    CM_PER_100M = 7
    H_O_CLOCK = 8

    INCH = 10
    FOOT = 11
    YARD = 12
    MILE = 13
    NAUTICAL_MILE = 14
    MILLIMETER = 15
    CENTIMETER = 16
    METER = 17
    KILOMETER = 18
    LINE = 19

    FOOT_POUND = 30
    JOULE = 31

    MM_HG = 40
    IN_HG = 41
    BAR = 42
    HP = 43
    PSI = 44

    FAHRENHEIT = 50
    CELSIUS = 51
    KELVIN = 52
    RANKIN = 53

    MPS = 60
    KMH = 61
    FPS = 62
    MPH = 63
    KT = 64

    GRAIN = 70
    OUNCE = 71
    GRAM = 72
    POUND = 73
    KILOGRAM = 74
    NEWTON = 75

    @property
    def key(self) -> str:
        """
        :rtype: str
        :return: readable name of the unit of measure
        """
        return UnitPropsDict[self].name

    @property
    def accuracy(self) -> int:
        """
        :rtype: int
        :return: default accuracy of the unit of measure
        """
        return UnitPropsDict[self].accuracy

    @property
    def symbol(self):
        """
        :rtype: str
        :return: short symbol of the unit of measure in CI
        """
        return UnitPropsDict[self].symbol

    def __call__(self: 'Unit', value: [int, float, 'AbstractUnit']) -> 'AbstractUnit':
        """Creates new unit instance by dot syntax
        :param self: unit as Unit enum
        :param value: numeric value of the unit
        :return: AbstractUnit instance
        """
        if isinstance(value, AbstractUnit):
            return value << self
        if 0 <= self < 10:
            obj = Angular(value, self)
        elif 10 <= self < 20:
            obj = Distance(value, self)
        elif 30 <= self < 40:
            obj = Energy(value, self)
        elif 40 <= self < 50:
            obj = Pressure(value, self)
        elif 50 <= self < 60:
            obj = Temperature(value, self)
        elif 60 <= self < 70:
            obj = Velocity(value, self)
        elif 70 <= self < 80:
            obj = Weight(value, self)
        else:
            raise TypeError(f"{self} Unit is not supports")
        return obj


class UnitProps(NamedTuple):
    """Properties of unit measure"""
    name: str
    accuracy: int
    symbol: str


UnitPropsDict = {
    Unit.RAD: UnitProps('radian', 6, 'rad'),
    Unit.DEGREE: UnitProps('degree', 4, '°'),
    Unit.MOA: UnitProps('MOA', 2, 'MOA'),
    Unit.MIL: UnitProps('MIL', 2, 'MIL'),
    Unit.MRAD: UnitProps('MRAD', 2, 'MRAD'),
    Unit.THOUSAND: UnitProps('thousand', 2, 'ths'),
    Unit.INCHES_PER_100YD: UnitProps('inches/100yd', 2, 'in/100yd'),
    Unit.CM_PER_100M: UnitProps('cm/100m', 2, 'cm/100m'),
    Unit.H_O_CLOCK: UnitProps('hour', 2, 'h'),

    Unit.INCH: UnitProps("inch", 3, "inch"),
    Unit.FOOT: UnitProps("foot", 2, "ft"),
    Unit.YARD: UnitProps("yard", 3, "yd"),
    Unit.MILE: UnitProps("mile", 3, "mi"),
    Unit.NAUTICAL_MILE: UnitProps("nautical mile", 3, "nm"),
    Unit.MILLIMETER: UnitProps("millimeter", 3, "mm"),
    Unit.CENTIMETER: UnitProps("centimeter", 3, "cm"),
    Unit.METER: UnitProps("meter", 3, "m"),
    Unit.KILOMETER: UnitProps("kilometer", 3, "km"),
    Unit.LINE: UnitProps("line", 3, "ln"),

    Unit.FOOT_POUND: UnitProps('foot * pound', 0, 'ft·lb'),
    Unit.JOULE: UnitProps('joule', 0, 'J'),

    Unit.MM_HG: UnitProps('mmHg', 0, 'mmHg'),
    Unit.IN_HG: UnitProps('inHg', 6, 'inHg'),
    Unit.BAR: UnitProps('bar', 2, 'bar'),
    Unit.HP: UnitProps('hPa', 4, 'hPa'),
    Unit.PSI: UnitProps('psi', 4, 'psi'),

    Unit.FAHRENHEIT: UnitProps('fahrenheit', 1, '°F'),
    Unit.CELSIUS: UnitProps('celsius', 1, '°C'),
    Unit.KELVIN: UnitProps('kelvin', 1, '°K'),
    Unit.RANKIN: UnitProps('rankin', 1, '°R'),

    Unit.MPS: UnitProps('mps', 0, 'm/s'),
    Unit.KMH: UnitProps('kmh', 1, 'km/h'),
    Unit.FPS: UnitProps('fps', 1, 'ft/s'),
    Unit.MPH: UnitProps('mph', 1, 'mph'),
    Unit.KT: UnitProps('knots', 1, 'kt'),

    Unit.GRAIN: UnitProps('grain', 1, 'gr'),
    Unit.OUNCE: UnitProps('ounce', 1, 'oz'),
    Unit.GRAM: UnitProps('gram', 1, 'g'),
    Unit.POUND: UnitProps('pound', 3, 'lb'),
    Unit.KILOGRAM: UnitProps('kilogram', 3, 'kg'),
    Unit.NEWTON: UnitProps('newton', 3, 'N'),
}


class AbstractUnit:
    """Abstract class for unit of measure instance definition
    Stores defined unit and value, applies conversions to other units
    """
    __slots__ = ('_value', '_defined_units')

    def __init__(self, value: [float, int], units: Unit):
        """
        :param units: unit as Unit enum
        :param value: numeric value of the unit
        """
        self._value: float = self.to_raw(value, units)
        self._defined_units: Unit = units

    def __str__(self) -> str:
        """Returns readable unit value
        :return: readable unit value
        """
        units = self._defined_units
        props = UnitPropsDict[units]
        v = self.from_raw(self._value, units)
        return f'{round(v, props.accuracy)}{props.symbol}'

    def __repr__(self):
        """Returns instance as readable view
        :return: instance as readable view
        """
        return f'<{self.__class__.__name__}: {self << self.units} ({round(self._value, 4)})>'

    def __float__(self):
        return float(self._value)

    def __eq__(self, other):
        return float(self) == other

    def __lt__(self, other):
        return float(self) < other

    def __gt__(self, other):
        return float(self) > other

    def __le__(self, other):
        return float(self) <= other

    def __ge__(self, other):
        return float(self) >= other

    def __lshift__(self, other: Unit):
        return self.convert(other)

    def __rshift__(self, other: Unit):
        return self.get_in(other)

    def __rlshift__(self, other: Unit):
        return self.convert(other)

    def _unit_support_error(self, value: float, units: Unit):
        """Validates the units
        :param value: value of the unit
        :param units: Unit enum type
        :return: value in specified units
        """
        if not isinstance(units, Unit):
            err_msg = f"Type expected: {Unit}, {type(Unit).__name__} " \
                      f"found: {type(units).__name__} ({value})"
            raise TypeError(err_msg)
        if units not in self.__dict__.values():
            raise ValueError(f'{self.__class__.__name__}: unit {units} is not supported')
        return 0

    def to_raw(self, value: float, units: Unit) -> float:
        """Converts value with specified units to raw value
        :param value: value of the unit
        :param units: Unit enum type
        :return: value in specified units
        """
        return self._unit_support_error(value, units)

    def from_raw(self, value: float, units: Unit) -> float:
        """Converts raw value to specified units
        :param value: raw value of the unit
        :param units: Unit enum type
        :return: value in specified units
        """
        return self._unit_support_error(value, units)

    def convert(self, units: Unit) -> 'AbstractUnit':
        """Returns new unit instance in specified units
        :param units: Unit enum type
        :return: new unit instance in specified units
        """
        value = self.get_in(units)
        return self.__class__(value, units)

    def get_in(self, units: Unit) -> float:
        """Returns value in specified units
        :param units: Unit enum type
        :return: value in specified units
        """
        return self.from_raw(self._value, units)

    @property
    def units(self) -> Unit:
        """Returns defined units
        :return: defined units
        """
        return self._defined_units

    @property
    def raw_value(self) -> float:
        """Raw unit value getter
        :return: raw unit value
        """
        return self._value


class Distance(AbstractUnit):
    """Distance unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Distance.Inch:
            return value
        if units == Distance.Foot:
            result = value * 12
        elif units == Distance.Yard:
            result = value * 36
        elif units == Distance.Mile:
            result = value * 63360
        elif units == Distance.NauticalMile:
            result = value * 72913.3858
        elif units == Distance.Line:
            result = value / 10
        elif units == Distance.Millimeter:
            result = value / 25.4
        elif units == Distance.Centimeter:
            result = value / 2.54
        elif units == Distance.Meter:
            result = value / 25.4 * 1000
        elif units == Distance.Kilometer:
            result = value / 25.4 * 1000000
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Distance.Inch:
            return value
        if units == Distance.Foot:
            result = value / 12
        elif units == Distance.Yard:
            result = value / 36
        elif units == Distance.Mile:
            result = value / 63360
        elif units == Distance.NauticalMile:
            result = value / 72913.3858
        elif units == Distance.Line:
            result = value * 10
        elif units == Distance.Millimeter:
            result = value * 25.4
        elif units == Distance.Centimeter:
            result = value * 2.54
        elif units == Distance.Meter:
            result = value * 25.4 / 1000
        elif units == Distance.Kilometer:
            result = value * 25.4 / 1000000
        else:
            return super().from_raw(value, units)
        return result

    Inch = Unit.INCH
    Foot = Unit.FOOT
    Yard = Unit.YARD
    Mile = Unit.MILE
    NauticalMile = Unit.NAUTICAL_MILE
    Millimeter = Unit.MILLIMETER
    Centimeter = Unit.CENTIMETER
    Meter = Unit.METER
    Kilometer = Unit.KILOMETER
    Line = Unit.LINE


class Pressure(AbstractUnit):
    """Pressure unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Pressure.MmHg:
            return value
        if units == Pressure.InHg:
            result = value * 25.4
        elif units == Pressure.Bar:
            result = value * 750.061683
        elif units == Pressure.HP:
            result = value * 750.061683 / 1000
        elif units == Pressure.PSI:
            result = value * 51.714924102396
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Pressure.MmHg:
            return value
        if units == Pressure.InHg:
            result = value / 25.4
        elif units == Pressure.Bar:
            result = value / 750.061683
        elif units == Pressure.HP:
            result = value / 750.061683 * 1000
        elif units == Pressure.PSI:
            result = value / 51.714924102396
        else:
            return super().from_raw(value, units)
        return result

    MmHg = Unit.MM_HG
    InHg = Unit.IN_HG
    Bar = Unit.BAR
    HP = Unit.HP
    PSI = Unit.PSI


class Weight(AbstractUnit):
    """Weight unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Weight.Grain:
            return value
        if units == Weight.Gram:
            result = value * 15.4323584
        elif units == Weight.Kilogram:
            result = value * 15432.3584
        elif units == Weight.Newton:
            result = value * 151339.73750336
        elif units == Weight.Pound:
            result = value / 0.000142857143
        elif units == Weight.Ounce:
            result = value * 437.5
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Weight.Grain:
            return value
        if units == Weight.Gram:
            result = value / 15.4323584
        elif units == Weight.Kilogram:
            result = value / 15432.3584
        elif units == Weight.Newton:
            result = value / 151339.73750336
        elif units == Weight.Pound:
            result = value * 0.000142857143
        elif units == Weight.Ounce:
            result = value / 437.5
        else:
            return super().from_raw(value, units)
        return result

    Grain = Unit.GRAIN
    Ounce = Unit.OUNCE
    Gram = Unit.GRAM
    Pound = Unit.POUND
    Kilogram = Unit.KILOGRAM
    Newton = Unit.NEWTON


class Temperature(AbstractUnit):
    """Temperature unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Temperature.Fahrenheit:
            return value
        if units == Temperature.Rankin:
            result = value - 459.67
        elif units == Temperature.Celsius:
            result = value * 9 / 5 + 32
        elif units == Temperature.Kelvin:
            result = (value - 273.15) * 9 / 5 + 32
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Temperature.Fahrenheit:
            return value
        if units == Temperature.Rankin:
            result = value + 459.67
        elif units == Temperature.Celsius:
            result = (value - 32) * 5 / 9
        elif units == Temperature.Kelvin:
            result = (value - 32) * 5 / 9 + 273.15
        else:
            return super().from_raw(value, units)
        return result

    Fahrenheit = Unit.FAHRENHEIT
    Celsius = Unit.CELSIUS
    Kelvin = Unit.KELVIN
    Rankin = Unit.RANKIN


class Angular(AbstractUnit):
    """Angular unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Angular.Radian:
            return value
        if units == Angular.Degree:
            result = value / 180 * pi
        elif units == Angular.MOA:
            result = value / 180 * pi / 60
        elif units == Angular.Mil:
            result = value / 3200 * pi
        elif units == Angular.MRad:
            result = value / 1000
        elif units == Angular.Thousand:
            result = value / 3000 * pi
        elif units == Angular.InchesPer100Yd:
            result = atan(value / 3600)
        elif units == Angular.CmPer100M:
            result = atan(value / 10000)
        elif units == Angular.OClock:
            result = value / 6 * pi
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Angular.Radian:
            return value
        if units == Angular.Degree:
            result = value * 180 / pi
        elif units == Angular.MOA:
            result = value * 180 / pi * 60
        elif units == Angular.Mil:
            result = value * 3200 / pi
        elif units == Angular.MRad:
            result = value * 1000
        elif units == Angular.Thousand:
            result = value * 3000 / pi
        elif units == Angular.InchesPer100Yd:
            result = tan(value) * 3600
        elif units == Angular.CmPer100M:
            result = tan(value) * 10000
        elif units == Angular.OClock:
            result = value * 6 / pi
        else:
            return super().from_raw(value, units)
        return result

    Radian = Unit.RAD
    Degree = Unit.DEGREE
    MOA = Unit.MOA
    Mil = Unit.MIL
    MRad = Unit.MRAD
    Thousand = Unit.THOUSAND
    InchesPer100Yd = Unit.INCHES_PER_100YD
    CmPer100M = Unit.CM_PER_100M
    OClock = Unit.H_O_CLOCK


class Velocity(AbstractUnit):
    """Velocity unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Velocity.MPS:
            return value
        if units == Velocity.KMH:
            return value / 3.6
        if units == Velocity.FPS:
            return value / 3.2808399
        if units == Velocity.MPH:
            return value / 2.23693629
        if units == Velocity.KT:
            return value / 1.94384449
        return super().to_raw(value, units)

    def from_raw(self, value: float, units: Unit):
        if units == Velocity.MPS:
            return value
        if units == Velocity.KMH:
            return value * 3.6
        if units == Velocity.FPS:
            return value * 3.2808399
        if units == Velocity.MPH:
            return value * 2.23693629
        if units == Velocity.KT:
            return value * 1.94384449
        return super().from_raw(value, units)

    MPS = Unit.MPS
    KMH = Unit.KMH
    FPS = Unit.FPS
    MPH = Unit.MPH
    KT = Unit.KT


class Energy(AbstractUnit):
    """Energy unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Energy.FootPound:
            return value
        if units == Energy.Joule:
            return value * 0.737562149277
        return super().to_raw(value, units)

    def from_raw(self, value: float, units: Unit):
        if units == Energy.FootPound:
            return value
        if units == Energy.Joule:
            return value / 0.737562149277
        return super().from_raw(value, units)

    FootPound = Unit.FOOT_POUND
    Joule = Unit.JOULE


@dataclass
class TypedUnits:  # pylint: disable=too-few-public-methods
    """
    Abstract class to apply auto-conversion values to
    specified units by type-hints in inherited dataclasses
    """

    def __setattr__(self, key, value):
        """
        converts value to specified units by type-hints in inherited dataclass
        """

        _fields = self.__getattribute__('__dataclass_fields__')
        # fields(self.__class__)[0].name
        if key in _fields and not isinstance(value, AbstractUnit):
            default_factory = _fields[key].default_factory
            if isinstance(default_factory, typing.Callable):
                if isinstance(value, Unit):
                    value = None
                else:
                    value = default_factory()(value)

        super().__setattr__(key, value)


# def is_unit(obj: [AbstractUnit, float, int]):
#     """
#     Check if obj is inherited by AbstractUnit
#     :return: False - if float or int
#     """
#     if isinstance(obj, AbstractUnit):
#         return True
#     if isinstance(obj, (float, int)):
#         return False
#     if obj is None:
#         return None
#     raise TypeError(f"Expected Unit, int, or float, found {obj.__class__.__name__}")


# Default units
# Angular.Radian
# Distance.Inch
# Energy.FootPound
# Weight.Grain
# Velocity.MPS
# Temperature.Fahrenheit
# Pressure.MmHg



"""LGPL library for small arms ballistic calculations (Python 3.8+)"""

__author__ = "o-murphy"
__copyright__ = ("",)

__credits__ = ["o-murphy"]
__version__ = "1.1.3"


from .backend import *
from .drag_tables import *
from .settings import *
from .multiple_bc import *
from .interface import *
from .trajectory_data import *
from .conditions import *
from .munition import *
from .unit import *



"""Classes to define zeroing or current environment conditions"""

import math
from dataclasses import dataclass, field

from .settings import Settings as Set
from .unit import Distance, Velocity, Temperature, Pressure, TypedUnits, Angular

__all__ = ('Atmo', 'Wind', 'Shot')

cIcaoStandardTemperatureR: float = 518.67
cIcaoFreezingPointTemperatureR: float = 459.67
cTemperatureGradient: float = -3.56616e-03
cIcaoStandardHumidity: float = 0.0
cPressureExponent: float = -5.255876
cSpeedOfSound: float = 49.0223
cA0: float = 1.24871
cA1: float = 0.0988438
cA2: float = 0.00152907
cA3: float = -3.07031e-06
cA4: float = 4.21329e-07
cA5: float = 3.342e-04
cStandardTemperature: float = 59.0
cStandardPressure: float = 29.92
cStandardDensity: float = 0.076474

cIcaoTemperatureDeltaR: float = cIcaoStandardTemperatureR - cIcaoFreezingPointTemperatureR


@dataclass
class Atmo(TypedUnits):  # pylint: disable=too-many-instance-attributes
    """Stores atmosphere data for the trajectory calculation"""

    altitude: [float, Distance] = field(default_factory=lambda: Set.Units.distance)
    pressure: [float, Pressure] = field(default_factory=lambda: Set.Units.pressure)
    temperature: [float, Temperature] = field(default_factory=lambda: Set.Units.temperature)
    humidity: float = 0.78
    density: float = field(init=False)
    mach: Velocity = field(init=False)
    _mach1: Velocity = field(init=False)
    _a0: float = field(init=False)
    _t0: float = field(init=False)
    _p0: float = field(init=False)
    _ta: float = field(init=False)

    def __post_init__(self):

        if self.humidity > 1:
            self.humidity = self.humidity / 100
        if not 0 <= self.humidity <= 1:
            self.humidity = 0.78
        if not self.altitude:
            self.altitude = Distance.Foot(0)
        if not self.pressure:
            self.pressure = Pressure.InHg(cStandardPressure)
        if not self.temperature:
            self.temperature = Temperature.Fahrenheit(cStandardTemperature)

        self.calculate()

    @staticmethod
    def icao(altitude: [float, Distance] = 0):
        """Creates Atmosphere with ICAO values"""
        altitude = Set.Units.distance(altitude)
        temperature = Temperature.Fahrenheit(
            cIcaoStandardTemperatureR + (altitude >> Distance.Foot)
            * cTemperatureGradient - cIcaoFreezingPointTemperatureR
        )

        pressure = Pressure.InHg(
            cStandardPressure * math.pow(cIcaoStandardTemperatureR / (
                    (temperature >> Temperature.Fahrenheit) + cIcaoFreezingPointTemperatureR),
                                    cPressureExponent
                                    )
        )

        return Atmo(
            altitude >> Set.Units.distance,
            pressure >> Set.Units.pressure,
            temperature >> Set.Units.temperature,
            cIcaoStandardHumidity
        )

    def density_factor(self):
        """:return: projectile density_factor"""
        return self.density / cStandardDensity

    def calculate0(self, t, p) -> (float, float):
        """:return: density and mach with specified atmosphere"""
        if t > 0:
            et0 = cA0 + t * (cA1 + t * (cA2 + t * (cA3 + t * cA4)))
            et = cA5 * self.humidity * et0
            hc = (p - 0.3783 * et) / cStandardPressure
        else:
            hc = 1.0

        density = cStandardDensity * (
                cIcaoStandardTemperatureR / (t + cIcaoFreezingPointTemperatureR)
        ) * hc
        mach = math.sqrt(t + cIcaoFreezingPointTemperatureR) * cSpeedOfSound
        return density, mach

    def calculate(self) -> None:
        """prepare the data for the calculation"""
        self._t0 = self.temperature >> Temperature.Fahrenheit
        self._p0 = self.pressure >> Pressure.InHg
        self._a0 = self.altitude >> Distance.Foot
        self._ta = self._a0 * cTemperatureGradient + cIcaoTemperatureDeltaR

        self.density, self._mach1 = self.calculate0(self._t0, self._p0)
        self.mach = Velocity(self._mach1, Velocity.FPS)

    def get_density_factor_and_mach_for_altitude(self, altitude: float):
        """:return: density factor for the specified altitude"""
        if math.fabs(self._a0 - altitude) < 30:
            density = self.density / cStandardDensity
            mach = self._mach1
            return density, mach

        tb = altitude * cTemperatureGradient + cIcaoTemperatureDeltaR
        t = self._t0 + self._ta - tb
        p = self._p0 * math.pow(self._t0 / t, cPressureExponent)

        density, mach = self.calculate0(t, p)
        return density / cStandardDensity, mach


@dataclass
class Wind(TypedUnits):
    """Stores wind data at the desired distance"""
    velocity: [float, Velocity] = field(default_factory=lambda: Set.Units.velocity)
    direction_from: [float, Angular] = field(default_factory=lambda: Set.Units.angular)
    until_distance: [float, Distance] = field(default_factory=lambda: Set.Units.distance)

    def __post_init__(self):
        if not self.until_distance:
            self.until_distance = Distance.Meter(9999)
        if not self.direction_from or not self.velocity:
            self.direction_from = 0
            self.velocity = 0


@dataclass
class Shot(TypedUnits):
    """
    Stores shot parameters for the trajectory calculation
    
    :param max_range: Downrange distance to stop computing trajectory
    :param zero_angle: The angle between the barrel and horizontal when zeroed
    :param relative_angle: Elevation adjustment added to zero_angle for a particular shot
    :param cant_angle: Rotation of gun around barrel axis, relative to position when zeroed.
        (Only relevant when Weapon.sight_height != 0)
    """
    max_range: [float, Distance] = field(default_factory=lambda: Set.Units.distance)
    zero_angle: [float, Angular] = field(default_factory=lambda: Set.Units.angular)
    relative_angle: [float, Angular] = field(default_factory=lambda: Set.Units.angular)
    cant_angle: [float, Angular] = field(default_factory=lambda: Set.Units.angular)

    atmo: Atmo = field(default=None)
    winds: list[Wind] = field(default=None)

    def __post_init__(self):
        if not self.relative_angle:
            self.relative_angle = 0
        if not self.cant_angle:
            self.cant_angle = 0
        if not self.atmo:
            self.atmo = Atmo.icao()
        if not self.winds:
            self.winds = [Wind()]



"""Module for Weapon and Ammo properties definitions"""
import math
from dataclasses import dataclass, field

from .settings import Settings as Set
from .unit import TypedUnits, Velocity, Temperature, Distance, Angular

__all__ = ('Weapon', 'Ammo')


@dataclass
class Weapon(TypedUnits):
    """
    :param zero_distance: Sight-line distance to "zero," which is point we want to hit.
        This is the distance that a rangefinder would return with no ballistic adjustment.
        NB: Some rangefinders offer an adjusted distance based on inclinometer measurement.
            However, without a complete ballistic model these can only approximate the effects
            on ballistic trajectory of shooting uphill or downhill.  Therefore:
            For maximum accuracy, use the raw sight distance and look angle as inputs here.
    :param zero_look_angle: If = 0 then the target zero is horizontal with the gun at zero_distance.
        If zero_look_angle != 0 then the target zero is at a different height than the gun, and
            the horizontal distance to target = cos(look_angle) * zero_distance, and
            the vertical distance to target from horizontal = sin(look_angle) * zero_distance.
    """
    sight_height: [float, Distance] = field(default_factory=lambda: Set.Units.sight_height)
    zero_distance: [float, Distance] = field(default_factory=lambda: Set.Units.distance)
    twist: [float, Distance] = field(default_factory=lambda: Set.Units.twist)
    zero_look_angle: [float, Angular] = field(default_factory=lambda: Set.Units.angular)

    def __post_init__(self):
        if not self.twist:
            self.twist = 0
        if not self.zero_look_angle:
            self.zero_look_angle = 0


@dataclass
class Ammo(TypedUnits):
    """Creates Ammo and Projectile properties"""

    dm: 'DragModel'
    length: [float, Distance] = field(default_factory=lambda: Set.Units.length)
    mv: [float, Velocity] = field(default_factory=lambda: Set.Units.velocity)
    temp_modifier: float = field(default=0)
    powder_temp: [float, Temperature] = field(default_factory=lambda: Temperature.Celsius(15))

    def calc_powder_sens(self, other_velocity: [float, Velocity],
                         other_temperature: [float, Temperature]) -> float:
        """Calculates velocity correction by temperature change
        :param other_velocity: other velocity
        :param other_temperature: other temperature
        :return: temperature modifier
        """
        # (800-792) / (15 - 0) * (15/792) * 100 = 1.01
        # creates temperature modifier in percent at each 15C
        v0 = self.mv >> Velocity.MPS
        t0 = self.powder_temp >> Temperature.Celsius
        v1 = Set.Units.velocity(other_velocity) >> Velocity.MPS
        t1 = Set.Units.temperature(other_temperature) >> Temperature.Celsius

        v_delta = math.fabs(v0 - v1)
        t_delta = math.fabs(t0 - t1)
        v_lower = v1 if v1 < v0 else v0

        if v_delta == 0 or t_delta == 0:
            raise ValueError(
                "Temperature modifier error, other velocity "
                "and temperature can't be same as default"
            )

        self.temp_modifier = v_delta / t_delta * (15 / v_lower)  # * 100

        return self.temp_modifier

    def get_velocity_for_temp(self, current_temp: [float, Temperature]) -> Velocity:
        """Calculates current velocity by temperature correction
        :param current_temp: temperature on current atmosphere
        :return: velocity corrected for temperature specified
        """
        temp_modifier = self.temp_modifier
        v0 = self.mv >> Velocity.MPS
        t0 = self.powder_temp >> Temperature.Celsius
        t1 = Set.Units.temperature(current_temp) >> Temperature.Celsius

        t_delta = t1 - t0
        muzzle_velocity = temp_modifier / (15 / v0) * t_delta + v0

        return Velocity.MPS(muzzle_velocity)



import unittest
from py_ballisticcalc import *


class TestDangerSpace(unittest.TestCase):

    def setUp(self) -> None:
        self.look_angle = Angular.Degree(0)
        weight, diameter = 168, 0.308
        length = Distance.Inch(1.282)
        weapon = Weapon(0, Distance.Foot(300), zero_look_angle=self.look_angle)
        dm = DragModel(0.223, TableG7, weight, diameter)
        ammo = Ammo(dm, length, Velocity.FPS(2750), Temperature.Celsius(15))
        ammo.calc_powder_sens(2723, 0)
        zero_atmo = Atmo.icao(100)
        calc = Calculator(weapon, ammo, zero_atmo)
        current_atmo = Atmo(110, 1000, 15, 72)
        current_winds = [Wind(2, 90)]
        shot = Shot(3000, atmo=current_atmo, winds=current_winds)
        self.shot_result = calc.fire(shot, Distance.Yard(100), extra_data=True)

    def test_danger_space(self):
        danger_space = self.shot_result.danger_space(
            Distance.Yard(500), Distance.Meter(1.5), self.look_angle
        )

        self.assertAlmostEqual(
            round(danger_space.begin.distance >> Distance.Yard, Distance.Yard.accuracy), 392.833, 2)
        self.assertAlmostEqual(
            round(danger_space.end.distance >> Distance.Yard, Distance.Yard.accuracy), 579.499, 2)

        danger_space = self.shot_result.danger_space(
            Distance.Yard(500), Distance.Inch(10), self.look_angle
        )

        self.assertAlmostEqual(
            round(danger_space.begin.distance >> Distance.Yard, Distance.Yard.accuracy), 484.367, 2)
        self.assertAlmostEqual(
            round(danger_space.end.distance >> Distance.Yard, Distance.Yard.accuracy), 514.967, 2)



import unittest
from py_ballisticcalc import *


class TestInterface(unittest.TestCase):

    def setUp(self) -> None:
        dm = DragModel(0.22, TableG7, 168, 0.308)
        self.ammo = Ammo(dm, 1.22, Velocity(2600, Velocity.FPS))
        self.atmosphere = Atmo.icao()



"""Unittests for the py_ballisticcalc library"""

import unittest
from math import fabs
from py_ballisticcalc import *


class TestTrajectory(unittest.TestCase):

    def test_zero1(self):
        dm = DragModel(0.365, TableG1, 69, 0.223)
        ammo = Ammo(dm, 0.9, 2600)
        weapon = Weapon(Distance(3.2, Distance.Inch), Distance(100, Distance.Yard))
        atmosphere = Atmo.icao()
        calc = TrajectoryCalc(ammo)

        zero_angle = calc.zero_angle(weapon, atmosphere)

        self.assertAlmostEqual(zero_angle >> Angular.Radian, 0.001651, 6,
                               f'TestZero1 failed {zero_angle >> Angular.Radian:.10f}')

    def test_zero2(self):
        dm = DragModel(0.223, TableG7, 69, 0.223)
        ammo = Ammo(dm, 0.9, 2750)
        weapon = Weapon(Distance(2, Distance.Inch), Distance(100, Distance.Yard))
        atmosphere = Atmo.icao()
        calc = TrajectoryCalc(ammo)

        zero_angle = calc.zero_angle(weapon, atmosphere)

        self.assertAlmostEqual(zero_angle >> Angular.Radian, 0.001228, 6,
                               f'TestZero2 failed {zero_angle >> Angular.Radian:.10f}')

    def custom_assert_equal(self, a, b, accuracy, name):
        with self.subTest(name=name):
            self.assertLess(fabs(a - b), accuracy, f'Assertion {name} failed ({a}/{b}, {accuracy})')

    def validate_one(self, data: TrajectoryData, distance: float, velocity: float,
                     mach: float, energy: float, path: float, hold: float,
                     windage: float, wind_adjustment: float, time: float, ogv: float,
                     adjustment_unit: Unit):

        self.custom_assert_equal(distance, data.distance >> Distance.Yard, 0.001, "Distance")
        self.custom_assert_equal(velocity, data.velocity >> Velocity.FPS, 5, "Velocity")
        self.custom_assert_equal(mach, data.mach, 0.005, "Mach")
        self.custom_assert_equal(energy, data.energy >> Energy.FootPound, 5, "Energy")
        self.custom_assert_equal(time, data.time, 0.06, "Time")
        self.custom_assert_equal(ogv, data.ogw >> Weight.Pound, 1, "OGV")

        if distance >= 800:
            self.custom_assert_equal(path, data.drop >> Distance.Inch, 4, 'Drop')
        elif distance >= 500:
            self.custom_assert_equal(path, data.drop >> Distance.Inch, 1, 'Drop')
        else:
            self.custom_assert_equal(path, data.drop >> Distance.Inch, 0.5, 'Drop')

        if distance > 1:
            self.custom_assert_equal(hold, data.drop_adj >> adjustment_unit, 0.5, 'Hold')

        if distance >= 800:
            self.custom_assert_equal(windage, data.windage >> Distance.Inch, 1.5, "Windage")
        elif distance >= 500:
            self.custom_assert_equal(windage, data.windage >> Distance.Inch, 1, "Windage")
        else:
            self.custom_assert_equal(windage, data.windage >> Distance.Inch, 0.5, "Windage")

        if distance > 1:
            self.custom_assert_equal(wind_adjustment,
                                     data.windage_adj >> adjustment_unit, 0.5, "WAdj")

    def test_path_g1(self):
        dm = DragModel(0.223, TableG1, 168, 0.308)
        ammo = Ammo(dm, 1.282, Velocity(2750, Velocity.FPS))
        weapon = Weapon(Distance(2, Distance.Inch), Distance(100, Distance.Yard))
        atmosphere = Atmo.icao()
        shot_info = Shot(1000,
                         zero_angle=Angular(0.001228, Angular.Radian),
                         atmo=atmosphere,
                         winds=[Wind(Velocity(5, Velocity.MPH), Angular(10.5, Angular.OClock))])
        calc = TrajectoryCalc(ammo)
        data = calc.trajectory(weapon, shot_info, Distance.Yard(100))

        self.custom_assert_equal(len(data), 11, 0.1, "Length")

        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, Angular.MOA],
            [data[1], 100, 2351.2, 2.106, 2061, 0, 0, -0.6, -0.6, 0.118, 550, Angular.MOA],
            [data[5], 500, 1169.1, 1.047, 509.8, -87.9, -16.8, -19.5, -3.7, 0.857, 67, Angular.MOA],
            [data[10], 1000, 776.4, 0.695, 224.9, -823.9, -78.7, -87.5, -8.4, 2.495, 20, Angular.MOA]
        ]

        for i, d in enumerate(test_data):
            with self.subTest(f"validate one {i}"):
                self.validate_one(*d)

    def test_path_g7(self):
        dm = DragModel(0.223, TableG7, 168, 0.308)
        ammo = Ammo(dm, 1.282, Velocity(2750, Velocity.FPS))
        weapon = Weapon(2, 100, 11.24)
        shot_info = Shot(Distance.Yard(1000),
                         zero_angle=Angular.MOA(4.221),
                         winds=[Wind(Velocity(5, Velocity.MPH), -45)])

        calc = TrajectoryCalc(ammo)
        data = calc.trajectory(weapon, shot_info, Distance.Yard(100))

        self.custom_assert_equal(len(data), 11, 0.1, "Length")

        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, Angular.Mil],
            [data[1], 100, 2544.3, 2.279, 2416, 0, 0, -0.35, -0.09, 0.113, 698, Angular.Mil],
            [data[5], 500, 1810.7, 1.622, 1226, -56.3, -3.18, -9.96, -0.55, 0.673, 252, Angular.Mil],
            [data[10], 1000, 1081.3, 0.968, 442, -401.6, -11.32, -50.98, -1.44, 1.748, 55, Angular.Mil]
        ]

        for i, d in enumerate(test_data):
            with self.subTest(f"validate one {i}"):
                self.validate_one(*d)


if __name__ == '__main__':
    unittest.main()



"""definitions for a bullet's drag model calculation"""

import typing
from dataclasses import dataclass

import math

from .settings import Settings as Set
from .unit import Weight, Distance
from .drag_tables import DragTablesSet

__all__ = ('DragModel', 'make_data_points')


@dataclass
class DragDataPoint:

    CD: float
    Mach: float

    def __iter__(self):
        yield self.CD
        yield self.Mach

    def __repr__(self):
        return f"DragDataPoint(CD={self.CD}, Mach={self.Mach})"


class DragModel:

    def __init__(self, value: float,
                 drag_table: typing.Iterable,
                 weight: [float, Weight],
                 diameter: [float, Distance]):
        self.__post__init__(value, drag_table, weight, diameter)

    def __post__init__(self, value: float, drag_table, weight, diameter):
        table_len = len(drag_table)
        error = ''

        if table_len <= 0:
            error = 'Custom drag table must be longer than 0'
        elif value <= 0:
            error = 'Drag coefficient must be greater than zero'

        if error:
            raise ValueError(error)

        if drag_table in DragTablesSet:
            self.value = value
        elif table_len > 0:
            self.value = 1  # or 0.999
        else:
            raise ValueError('Wrong drag data')

        self.weight = Set.Units.weight(weight)
        self.diameter = Set.Units.diameter(diameter)
        self.sectional_density = self._get_sectional_density()
        self.form_factor = self._get_form_factor(self.value)
        self.drag_table = drag_table

    def _get_form_factor(self, bc: float):
        return self.sectional_density / bc

    def _get_sectional_density(self) -> float:
        w = self.weight >> Weight.Grain
        d = self.diameter >> Distance.Inch
        return sectional_density(w, d)

    @staticmethod
    def from_mbc(mbc: 'MultiBC'):
        return DragModel(1, mbc.cdm, mbc.weight, mbc.diameter)


def make_data_points(drag_table: typing.Iterable) -> list:
    return [DragDataPoint(point['CD'], point['Mach']) for point in drag_table]


def sectional_density(weight: float, diameter: float):
    return weight / math.pow(diameter, 2) / 7000



import unittest
from py_ballisticcalc import *


class TestMBC(unittest.TestCase):

    def test_mbc(self):
        mbc = MultiBC(
            drag_table=TableG7,
            weight=Weight(178, Weight.Grain),
            diameter=Distance(0.308, Distance.Inch),
            mbc_table=[{'BC': p[0], 'V': p[1]} for p in ((0.275, 800), (0.255, 500), (0.26, 700))],
        )
        dm = DragModel.from_mbc(mbc)
        ammo = Ammo(dm, 1, 800)
        cdm = TrajectoryCalc(ammo=ammo).cdm
        self.assertIsNot(cdm, None)
        ret = list(cdm)
        self.assertEqual(ret[0], {'Mach': 0.0, 'CD': 0.1259323091692403})
        self.assertEqual(ret[-1], {'Mach': 5.0, 'CD': 0.15771258594668947})

    def test_mbc_valid(self):
        # Litz's multi-bc table comversion to CDM, 338LM 285GR HORNADY ELD-M
        mbc = MultiBC(
            drag_table=TableG7,
            weight=Weight.Grain(285),
            diameter=Distance.Inch(0.338),
            mbc_table=[{'BC': p[0], 'V': Velocity.MPS(p[1])} for p in ((0.417, 745), (0.409, 662), (0.4, 580))],
        )
        cdm = mbc.cdm
        cds = [p['CD'] for p in cdm]
        machs = [p['Mach'] for p in cdm]

        reference = (
            (1, 0.3384895315),
            (2, 0.2573873416),
            (3, 0.2069547831),
            (4, 0.1652052415),
            (5, 0.1381406102),
        )

        for mach, cd in reference:
            idx = machs.index(mach)
            with self.subTest(mach=mach):
                self.assertAlmostEqual(cds[idx], cd, 3)



"""Implements a point of trajectory class in applicable data types"""
import logging
import math
import typing
from dataclasses import dataclass, field
from enum import Flag
from typing import NamedTuple

from .settings import Settings as Set
from .unit import Angular, Distance, Weight, Velocity, Energy, AbstractUnit, Unit
from .munition import Weapon
from .conditions import Shot

try:
    import pandas as pd
except ImportError as error:
    logging.warning("Install pandas to convert trajectory to dataframe")
    pd = None

try:
    import matplotlib
    from matplotlib import patches
except ImportError as error:
    logging.warning("Install matplotlib to get results as a plot")
    matplotlib = None

__all__ = ('TrajectoryData', 'HitResult', 'TrajFlag')


PLOT_FONT_HEIGHT = 72
PLOT_FONT_SIZE = 552 / PLOT_FONT_HEIGHT


class TrajFlag(Flag):
    """Flags for marking trajectory row if Zero or Mach crossing
    Also uses to set a filters for a trajectory calculation loop
    """
    NONE = 0
    ZERO_UP = 1
    ZERO_DOWN = 2
    MACH = 4
    RANGE = 8
    DANGER = 16
    ZERO = ZERO_UP | ZERO_DOWN
    ALL = RANGE | ZERO_UP | ZERO_DOWN | MACH | DANGER


class TrajectoryData(NamedTuple):
    """
    Represents point of trajectory in applicable data types

    Attributes:
        time (float): bullet flight time
        distance (Distance): traveled distance
        velocity (Velocity): velocity in current trajectory point
        mach (float): velocity in current trajectory point in "Mach" number
        drop (Distance):
        drop_adj (Angular):
        windage (Distance):
        windage_adj (Angular):
        angle (Angular)
        mach float
        energy (Energy):
        ogw (Weight): optimal game weight
        rtype (int): row type
    """

    time: float
    distance: Distance
    velocity: Velocity
    mach: float  # velocity in Mach
    drop: Distance
    drop_adj: Angular  # drop_adjustment
    windage: Distance
    windage_adj: Angular  # windage_adjustment
    angle: Angular  # Trajectory angle
    energy: Energy
    ogw: Weight
    flag: typing.Union[TrajFlag, int]

    def formatted(self) -> tuple:
        """
        :return: matrix of formatted strings for each value of trajectory in default units
        """

        def _fmt(v: AbstractUnit, u: Unit):
            """simple formatter"""
            return f"{v >> u:.{u.accuracy}f} {u.symbol}"

        return (
            f'{self.time:.2f} s',
            _fmt(self.distance, Set.Units.distance),
            _fmt(self.velocity, Set.Units.velocity),
            f'{self.mach:.2f} mach',
            _fmt(self.drop, Set.Units.drop),
            _fmt(self.drop_adj, Set.Units.adjustment),
            _fmt(self.windage, Set.Units.drop),
            _fmt(self.windage_adj, Set.Units.adjustment),
            _fmt(self.angle, Set.Units.angular),
            _fmt(self.energy, Set.Units.energy),
            _fmt(self.ogw, Set.Units.ogw),

            self.flag
        )

    def in_def_units(self) -> tuple:
        """
        :return: matrix of floats of the trajectory in default units
        """
        return (
            self.time,
            self.distance >> Set.Units.distance,
            self.velocity >> Set.Units.velocity,
            self.mach,
            self.drop >> Set.Units.drop,
            self.drop_adj >> Set.Units.adjustment,
            self.windage >> Set.Units.drop,
            self.windage_adj >> Set.Units.adjustment,
            self.angle >> Set.Units.angular,
            self.energy >> Set.Units.energy,
            self.ogw >> Set.Units.ogw,
            TrajFlag(self.flag)
        )


class DangerSpace(NamedTuple):
    """Stores the danger space data for distance specified"""
    at_range: TrajectoryData
    target_height: Distance
    begin: TrajectoryData
    end: TrajectoryData
    look_angle: Angular

    def __str__(self) -> str:
        return f'Danger space at {self.at_range.distance << Set.Units.distance} '\
             + f'for {self.target_height << Set.Units.drop} tall target '\
             + f'ranges from {self.begin.distance << Set.Units.distance} '\
             + f'to {self.end.distance << Set.Units.distance}'

    def overlay(self, ax: 'Axes'):
        """Highlights danger-space region on plot"""
        if matplotlib is None:
            raise ImportError("Install matplotlib to get results as a plot")

        begin_dist = self.begin.distance >> Set.Units.distance
        begin_drop = self.begin.drop >> Set.Units.drop
        end_dist = self.end.distance >> Set.Units.distance
        end_drop = self.end.drop >> Set.Units.drop
        range_dist = self.at_range.distance >> Set.Units.distance
        range_drop = self.at_range.drop >> Set.Units.drop
        h = self.target_height >> Set.Units.drop

        # Target position and height:
        ax.plot((range_dist, range_dist), (range_drop + h / 2, range_drop - h / 2),
                color='r', linestyle=':')
        # Shaded danger-space region:
        vertices = (
            (begin_dist, begin_drop), (end_dist, begin_drop),
            (end_dist, end_drop), (begin_dist, end_drop)
        )
        polygon = patches.Polygon(vertices, closed=True,
                                  edgecolor='none', facecolor='r', alpha=0.5)
        ax.add_patch(polygon)
        ax.text(begin_dist + (end_dist-begin_dist)/2, end_drop,
                f"Danger space\nat {self.at_range.distance << Set.Units.distance}",
                linespacing=1.2, fontsize=PLOT_FONT_SIZE, ha='center', va='top')


@dataclass(frozen=True)
class HitResult:
    """Results of the shot"""
    weapon: Weapon
    shot: Shot
    trajectory: list[TrajectoryData] = field(repr=False)
    extra: bool = False

    def __iter__(self):
        for row in self.trajectory:
            yield row

    def __getitem__(self, item):
        return self.trajectory[item]

    def __check_extra__(self):
        if not self.extra:
            raise AttributeError(
                f"{object.__repr__(self)} has no extra data. "
                f"Use Calculator.fire(..., extra_data=True)"
            )

    def zeros(self) -> list[TrajectoryData]:
        """:return: zero crossing points"""
        self.__check_extra__()
        data = [row for row in self.trajectory if row.flag & TrajFlag.ZERO.value]
        if len(data) < 1:
            raise ArithmeticError("Can't find zero crossing points")
        return data

    def danger_space(self,
                     at_range: [float, Distance],
                     target_height: [float, Distance],
                     look_angle: [float, Angular] = Angular(0, Angular.Degree)
                     ) -> DangerSpace:
        """
        Assume that the trajectory hits the center of a target at any distance.
        Now we want to know how much ranging error we can tolerate if the critical region 
        of the target has height *h*.  I.e., we want to know how far forward and backward
        along the line of sight we can move a target such that the trajectory is still 
        within *h*/2 of the original drop.

        :param at_range: Danger space is calculated for a target centered at this distance
        :param target_height: Target height (*h*) determines danger space
        :param look_angle: Ranging errors occur along the look angle to the target
        """
        self.__check_extra__()

        at_range = Set.Units.distance(at_range)
        target_height = Set.Units.distance(target_height)
        look_angle = Set.Units.angular(look_angle)

        # Get index of first trajectory point with distance >= at_range
        i = next((i for i in range(len(self.trajectory))
                  if self.trajectory[i].distance >= at_range), -1)
        if i < 0:
            raise ArithmeticError(
                f"Calculated trajectory doesn't reach requested distance {at_range}"
            )

        target_height_half = target_height.raw_value / 2.0
        tan_look_angle = math.tan(look_angle >> Angular.Radian)

        # Target_center height shifts along look_angle as:
        #   target_center' = target_center + (.distance' - .distance) * tan(look_angle)

        def find_begin_danger(row_num: int) -> TrajectoryData:
            """
            Beginning of danger space is last .distance' < .distance where 
                |target_center - .drop'| >= target_height/2
            :param row_num: Index of the trajectory point for which we are calculating danger space
            :return: Distance marking beginning of danger space
            """
            center_row = self.trajectory[row_num]
            for i in range(row_num - 1, 0, -1):
                prime_row = self.trajectory[i]
                target_center = center_row.drop.raw_value + tan_look_angle * (
                        prime_row.distance.raw_value - center_row.distance.raw_value
                )
                if abs(target_center - prime_row.drop.raw_value) >= target_height_half:
                    return self.trajectory[i]
            return self.trajectory[0]

        def find_end_danger(row_num: int) -> TrajectoryData:
            """
            End of danger space is first .distance' > .distance where 
                |target_center - .drop'| >= target_height/2
            :param row_num: Index of the trajectory point for which we are calculating danger space
            :return: Distance marking end of danger space
            """
            center_row = self.trajectory[row_num]
            for i in range(row_num + 1, len(self.trajectory)):
                prime_row = self.trajectory[i]
                target_center = center_row.drop.raw_value + tan_look_angle * (
                        prime_row.distance.raw_value - center_row.distance.raw_value)
                if abs(target_center - prime_row.drop.raw_value) >= target_height_half:
                    return prime_row
            return self.trajectory[-1]

        return DangerSpace(self.trajectory[i],
                           target_height,
                           find_begin_danger(i),
                           find_end_danger(i),
                           look_angle)

    @property
    def dataframe(self):
        """:return: the trajectory table as a DataFrame"""
        if pd is None:
            raise ImportError("Install pandas to get trajectory as dataframe")
        # self.__check_extra__()
        col_names = list(TrajectoryData._fields)
        trajectory = [p.in_def_units() for p in self]
        return pd.DataFrame(trajectory, columns=col_names)

    def plot(self, look_angle: Angular = None) -> 'Axes':
        """:return: graph of the trajectory"""
        if look_angle is None:
            look_angle = self.weapon.zero_look_angle

        if matplotlib is None:
            raise ImportError("Install matplotlib to plot results")
        if not self.extra:
            logging.warning("HitResult.plot: To show extended data"
                            "Use Calculator.fire(..., extra_data=True)")
        font_size = 552 / 72.0
        df = self.dataframe
        ax = df.plot(x='distance', y=['drop'], ylabel=Set.Units.drop.symbol)
        max_range = self.trajectory[-1].distance >> Set.Units.distance

        for p in self.trajectory:
            if TrajFlag(p.flag) & TrajFlag.ZERO:
                ax.plot([p.distance >> Set.Units.distance, p.distance >> Set.Units.distance],
                        [df['drop'].min(), p.drop >> Set.Units.drop], linestyle=':')
                ax.text((p.distance >> Set.Units.distance) + max_range/100, df['drop'].min(),
                        f"{(TrajFlag(p.flag) & TrajFlag.ZERO).name}",
                        fontsize=font_size, rotation=90)
            if TrajFlag(p.flag) & TrajFlag.MACH:
                ax.plot([p.distance >> Set.Units.distance, p.distance >> Set.Units.distance],
                        [df['drop'].min(), p.drop >> Set.Units.drop], linestyle=':')
                ax.text((p.distance >> Set.Units.distance) + max_range/100, df['drop'].min(), "Mach 1",
                        fontsize=font_size, rotation=90)

        max_range_in_drop_units = self.trajectory[-1].distance >> Set.Units.drop
        # Sight line
        x_sight = [0, df.distance.max()]
        y_sight = [0, max_range_in_drop_units * math.tan(look_angle >> Angular.Radian)]
        ax.plot(x_sight, y_sight, linestyle='--', color=[.3,0,.3,.5])
        # Barrel pointing line
        x_bbl = [0, df.distance.max()]
        y_bbl = [-(self.weapon.sight_height >> Set.Units.drop),
                    max_range_in_drop_units * math.tan(self.trajectory[0].angle >> Angular.Radian)
                    -(self.weapon.sight_height >> Set.Units.drop)]
        ax.plot(x_bbl, y_bbl, linestyle=':', color=[0,0,0,.5])
        # Line labels
        sight_above_bbl = True if y_sight[1] > y_bbl[1] else False
        angle = math.degrees(math.atan((y_sight[1]-y_sight[0])/(x_sight[1]-x_sight[0])))
        ax.text(x_sight[1], y_sight[1], "Sight line", linespacing=1.2,
                 rotation=angle, rotation_mode='anchor', transform_rotates_text=True,
                 fontsize=font_size, color=[.3,0,.3,1], ha='right',
                 va='bottom' if sight_above_bbl else 'top')
        angle = math.degrees(math.atan((y_bbl[1]-y_bbl[0])/(x_bbl[1]-x_bbl[0])))
        ax.text(x_bbl[1], y_bbl[1], "Barrel pointing", linespacing=1.2,
                 rotation=angle, rotation_mode='anchor', transform_rotates_text=True,
                 fontsize=font_size, color='k', ha='right',
                 va='top' if sight_above_bbl else 'bottom')

        df.plot(x='distance', xlabel=Set.Units.distance.symbol,
                y=['velocity'], ylabel=Set.Units.velocity.symbol,
                secondary_y=True, color=[0,.3,0,.5],
                ylim=[df['velocity'].min() - (Unit.FPS(100) >> Set.Units.velocity),
                      df['velocity'].max() + (Unit.FPS(100) >> Set.Units.velocity)],
                ax=ax)

        # Let secondary shine through
        ax.set_zorder(1)
        ax.set_facecolor([0,0,0,0])

        return ax



import unittest
from py_ballisticcalc import *


class TestPerformance(unittest.TestCase):
    def setUp(self) -> None:
        self.dm = DragModel(0.223, TableG7, 168, 0.308)
        self.ammo = Ammo(self.dm, 1.282, 2750)
        self.weapon = Weapon(2, 100, 11.24)
        self.atmo = Atmo.icao()
        self.shot = Shot(
            Distance.Yard(1000),
            zero_angle=Angular.MOA(4.221),
            winds=[Wind(Velocity(5, Velocity.MPH), -45)]
        )

        self.calc = TrajectoryCalc(self.ammo)

    def test__init__(self):
        self.assertTrue(self.calc)

    def test_elevation_performance(self):
        self.calc.zero_angle(self.weapon, self.atmo)

    def test_path_performance(self):
        d = self.calc.trajectory(self.weapon, self.shot, Distance.Yard(100))
        # [print(p.formatted()) for p in d]



"""Example of library usage"""


from py_ballisticcalc import *
from py_ballisticcalc import Settings as Set


# set global library settings
Set.Units.velocity = Velocity.FPS
Set.Units.temperature = Temperature.Celsius
# Set.Units.distance = Distance.Meter
Set.Units.sight_height = Distance.Centimeter

# Set.set_max_calc_step_size(Distance.Foot(1))
Set.USE_POWDER_SENSITIVITY = True  # enable muzzle velocity correction my powder temperature

# define params with default units
weight, diameter = 168, 0.308
# or define with specified units
length = Distance.Inch(1.282)  # length = Distance(1.282, Distance.Inch)

weapon = Weapon(9, 100, 2)
dm = DragModel(0.223, TableG7, weight, diameter)

ammo = Ammo(dm, length, 2750, 15)
ammo.calc_powder_sens(2723, 0)

zero_atmo = Atmo.icao(100)

# defining calculator instance
calc = Calculator(weapon, ammo, zero_atmo)

current_atmo = Atmo(110, 1000, 15, 72)
current_winds = [Wind(2, 90)]
shot = Shot(1500, atmo=current_atmo, winds=current_winds)

shot_result = calc.fire(shot, Distance.Yard(100))

for p in shot_result:
    print(p.formatted())



"""Implements basic interface for the ballistics calculator"""
from dataclasses import dataclass, field

from .conditions import Atmo, Shot
from .munition import Weapon, Ammo
# pylint: disable=import-error,no-name-in-module
from .backend import *
from .trajectory_data import HitResult
from .unit import Angular, Distance
from .settings import Settings

__all__ = ('Calculator',)


@dataclass
class Calculator:
    """Basic interface for the ballistics calculator"""

    weapon: Weapon
    ammo: Ammo
    zero_atmo: Atmo = field(default_factory=Atmo.icao)
    _elevation: Angular = field(init=False, repr=True, compare=False,
                                default_factory=lambda: Angular.Degree(0))
    _calc: TrajectoryCalc = field(init=False, repr=True, compare=False, default=None)

    def __post_init__(self):
        self.calculate_elevation()

    @property
    def elevation(self):
        """get current barrel elevation"""
        return self._elevation

    @property
    def cdm(self):
        """returns custom drag function based on input data"""
        return self._calc.cdm

    def calculate_elevation(self):
        """Recalculates barrel elevation for weapon and zero atmo"""
        self._calc = TrajectoryCalc(self.ammo)
        self._elevation = self._calc.zero_angle(self.weapon, self.zero_atmo)

    def fire(self, shot: Shot, trajectory_step: [float, Distance],
             extra_data: bool = False) -> HitResult:
        """Calculates trajectory
        :param shot: shot parameters
        :param trajectory_step: step between trajectory points
        :param extra_data: True => store TrajectoryData for every step;
            False => store TrajectoryData only for each trajectory_step
        """
        step = Settings.Units.distance(trajectory_step)
        self._calc = TrajectoryCalc(self.ammo)
        if not shot.zero_angle:
            shot.zero_angle = self._elevation
        data = self._calc.trajectory(self.weapon, shot, step, extra_data)
        return HitResult(self.weapon, shot, data, extra_data)



"""Global settings of the py_ballisticcalc library"""
import logging

from .unit import Unit, Distance

__all__ = ('Settings',)


class Settings:  # pylint: disable=too-few-public-methods
    """Global settings class of the py_ballisticcalc library"""

    class Units:  # pylint: disable=too-few-public-methods
        """Default units for specified measures"""

        sight_height: Unit = Unit.INCH
        twist: Unit = Unit.INCH
        velocity: Unit = Unit.FPS
        distance: Unit = Unit.YARD
        temperature: Unit = Unit.CELSIUS
        weight: Unit = Unit.GRAIN
        length: Unit = Unit.INCH
        diameter: Unit = Unit.INCH
        pressure: Unit = Unit.HP
        drop: Unit = Unit.CENTIMETER
        angular: Unit = Unit.DEGREE
        adjustment: Unit = Unit.MIL
        energy: Unit = Unit.JOULE
        ogw: Unit = Unit.POUND
        target_height: Unit = Unit.INCH

    _MAX_CALC_STEP_SIZE: float = 1
    USE_POWDER_SENSITIVITY: bool = False

    @classmethod
    def set_max_calc_step_size(cls, value: [float, Distance]):
        """_MAX_CALC_STEP_SIZE setter
        :param value: [float, Distance] maximum calculation step (used internally)
        """
        logging.warning("Settings._MAX_CALC_STEP_SIZE: change this property "
                        "only if you know what you are doing; "
                        "too big step can corrupt calculation accuracy")
        if not isinstance(value, (Distance, float, int)):
            raise ValueError("MIN_CALC_STEP_SIZE have to be a type of 'Distance'")
        cls._MAX_CALC_STEP_SIZE = cls.Units.distance(value) >> Distance.Foot



import unittest


class TestImports(unittest.TestCase):

    def test_bin_import(self):
        try:
            import py_ballisticcalc_exts
        except ImportError as err:
            print(err)
            py_ballisticcalc_exts = None
        if py_ballisticcalc_exts:
            from py_ballisticcalc.backend import DragModel
            print(DragModel)
